from . import soundpoint
soundpoint.setopt(randomize=True, duration=0.2)


def right_left_walk(size:int) -> [int]:
    """Yield index of objects to visit for a right left walk

    >>> tuple(right_left_walk(5))
    (2, 3, 1, 4, 0)

    """
    soundpoint()
    middle = size // 2
    yield middle
    low, up = middle-1, middle+1
    change = True
    while change:
        change = False
        if up < size:
            soundpoint()
            yield up
            up += 1
            change = True
        if low >= 0:
            soundpoint()
            yield low
            low -= 1
            change = True
    soundpoint()


if __name__ == '__main__':
    print('RUNNING…')
    tuple(right_left_walk(10))
    print('END')



"""
Adapted from pysine module.

On fedora, you need to install portaudio-devel: dnf install portaudio-devel

"""


import threading
from pyaudio import PyAudio
import numpy as np



class PySine:
    BITRATE = 96000

    def __init__(self):
        self.pyaudio = PyAudio()
        self.stream = self.pyaudio.open(
            format=self.pyaudio.get_format_from_width(1),
            channels=1,
            rate=self.BITRATE,
            output=True
        )

    def __del__(self):
        self.stream.stop_stream()
        self.stream.close()
        self.pyaudio.terminate()

    def sine(self, frequency=440.0, duration=1.0):
        points = int(self.BITRATE * duration)
        times = np.linspace(0, duration, points, endpoint=False)
        data = np.array((np.sin(times*frequency*2*np.pi) + 1.0)*127.5, dtype=np.int8).tostring()
        self.stream.write(data)


def play(frequency=440.0, duration=0.07, *, cls=PySine(), inthread:bool=False):
    call = lambda: cls.sine(frequency=frequency, duration=duration)
    if inthread:
        if play.thread:
            play.thread.join()
        play.thread = threading.Thread(target=call)
        play.thread.start()
    else:
        call()
play.thread = None


if __name__ == "__main__":
    play(440)



"""Definition of the Synepoint metaclass, allowing for further
extension of breakpoints behaviour.

"""

import traceback
import magwitch
from functools import wraps


# @magwitch.on  # TODO: make call method having the exact parameters
class Synepoint:
    """Metaclass for all breakpoints of syne package.

    A synepoint subclass is the representation of a type of synepoint,
    for instance a synepoint that beeps when encountered.
    Subclasses are converted as functions that user can import and use,
    and each call to these functions are identifiable in the class.

    """

    def __new__(cls, clsname, bases, attrs):

        def call_method(self, **kwargs):
            trace = traceback.format_stack()[-2]  # the last one, except this level and the call level of subclass
            h = hash(tuple(trace))
            if h not in self.synepoints:
                self.synepoints[h] = self.new_point()
            # call the function with options changed according to given kwargs
            options = dict(self.options)
            options.update(kwargs)
            self.options, options = options, self.options
            self.on_point(*self.synepoints[h])
            self.options, options = options, self.options  # restore
        def init_method(self):
            self.set_opt()

        attrs['__call__'] = call_method
        attrs['__init__'] = init_method
        attrs['synepoints'] = {}  # map synepoint -> payload data
        attrs['options'] = {}  # map option name -> option default

        # wrap set_opt with the standard assignation of self.options
        attrs['_set_opt'] = attrs['set_opt']
        @wraps(attrs['_set_opt'])
        def set_opt_method(self, **kwargs):
            self.options = self._set_opt(**kwargs)
        attrs['setopt'] = attrs['set_opt'] = set_opt_method

        # finally build the function and the class
        final_cls = type(clsname.title(), bases, attrs)
        globals()[clsname.lower()] = final_cls()  # associate an instance with the function
        return final_cls



"""Definition of soundpoint, based on synepoint.

"""
import random
import time
import musthe  # theoretical music
from .synepoint import Synepoint
from .utils import right_left_walk
from . import playsound


class Soundpoint(metaclass=Synepoint):
    """A soundpoint associate to one breakpoint a tone on a scale.

    """

    def new_point(self):
        "Return a new note/frequency"
        new_freq = next(self.options['notes'])
        return new_freq.scientific_notation(), new_freq.frequency()

    def on_point(self, name:str, freq:float):
        "Action to do when encountering a point with given payload"
        if self.options['show_notes']:
            print(self.options['show_notes_template'].format(name, freq))
        time_between_notes = self.options['time_between_notes']
        inthread = time_between_notes != 0
        playsound.play(freq, self.options['duration'], inthread=inthread)
        if time_between_notes is not None and time_between_notes > 0:
            time.sleep(time_between_notes)

    def set_opt(self, scale:str='B2 major',
                duration:float=0.08, show_notes:bool=False,
                randomize:bool=False, time_between_notes:float=None,
                show_notes_template:str='{name} ({freq})'):
        "Populate config based on given args"
        if scale in {None, 'all', 'ALL'}:
            # use all available notes
            scale = musthe.Note.all(min_octave=2, max_octave=6)  # not too high, it hurts
        else:  # valid input: use given scale
            scale_root, scale_type = scale.split(' ') if ' ' in scale else (scale, 'major')
            scale = musthe.Scale(scale_root.upper(), scale_type.lower())
        if randomize:
            scale = list(scale)
            random.shuffle(scale)
        else:  # sort by making very different notes succeeding each others
            scale = list(scale)
            scale = tuple(scale[idx] for idx in right_left_walk(len(scale)))
        scale = tuple(scale)
        return {
            'scale': scale,
            'notes': iter(scale),
            'duration': float(duration),
            'show_notes': bool(show_notes),
            'time_between_notes': time_between_notes,
            'show_notes_template': str(show_notes_template),
        }
